/*
 * ArgumentList.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using Eagle._Attributes;
using Eagle._Components.Private;
using Eagle._Components.Public;
using Eagle._Constants;
using Eagle._Interfaces.Public;

namespace Eagle._Containers.Public
{
#if SERIALIZATION
    [Serializable()]
#endif
    [ObjectId("65ed4894-90ef-49bf-8744-2a7c3406af55")]
    public sealed class ArgumentList : List<Argument>
    {
        #region Private Constructors
        private ArgumentList()
            : base()
        {
            // do nothing.
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Public Constructors
        public ArgumentList(
            params object[] objects
            )
            : base()
        {
            int length = objects.Length;

            if (length > 0)
            {
                for (int index = 0; index < length; index++)
                {
                    if (objects[index] is ArgumentList)
                    {
                        //
                        // NOTE: They supplied an argument list as a parameter.
                        //       Add the entire list to this list.
                        //
                        this.AddRange((ArgumentList)objects[index]);
                    }
                    else if (objects[index] is StringList)
                    {
                        //
                        // NOTE: They supplied a string list as a parameter.
                        //       Add the entire list to this list.
                        //
                        this.AddRange((StringList)objects[index]);
                    }
                    else
                    {
                        Argument argument;

                        if (objects[index] is Argument)
                        {
                            argument = (Argument)objects[index];
                        }
                        else if (objects[index] is string)
                        {
                            argument = new Argument((string)objects[index]);
                        }
                        else if (objects[index] != null)
                        {
                            argument = new Argument(objects[index].ToString());
                        }
                        else
                        {
                            argument = new Argument(Argument.NoValue);
                        }

                        this.Add(argument);
                    }
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////

        public ArgumentList(
            IEnumerable<Argument> collection
            )
            : base(collection)
        {
            // do nothing.
        }

        ///////////////////////////////////////////////////////////////////////

        public ArgumentList(
            IEnumerable<string> collection,
            ArgumentFlags flags
            )
            : this()
        {
            foreach (string item in collection)
            {
                Argument argument;

                if (FlagOps.HasFlags(flags, ArgumentFlags.NameOnly, true))
                    argument = new Argument(flags, item);
                else
                    argument = new Argument(flags, Argument.NoName, item);

                this.Add(argument);
            }
        }

        ///////////////////////////////////////////////////////////////////////

        public ArgumentList(
            IEnumerable<IPair<string>> collection,
            ArgumentFlags flags
            )
            : this()
        {
            foreach (IPair<string> item in collection)
            {
                ArgumentFlags valueFlags = (item.Y != null) ?
                    ArgumentFlags.HasDefault : ArgumentFlags.None;

                Argument argument;

                if (FlagOps.HasFlags(flags, ArgumentFlags.NameOnly, true))
                    argument = new Argument(
                        flags | valueFlags, item.X, Argument.NoValue, item.Y);
                else
                    argument = new Argument(
                        flags | valueFlags, Argument.NoName, item.X, item.Y);

                this.Add(argument);
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Public Static Methods
        public static ArgumentList GetRange(
            ArgumentList list,
            int firstIndex
            )
        {
            return GetRange(list, firstIndex,
                (list != null) ? (list.Count - 1) : Index.Invalid);
        }

        ///////////////////////////////////////////////////////////////////////

        public static ArgumentList GetRange(
            IList list,
            int firstIndex,
            int lastIndex
            )
        {
            ArgumentList range = null;

            if (list != null)
            {
                range = new ArgumentList();

                if (firstIndex == Index.Invalid)
                    firstIndex = 0;

                if (lastIndex == Index.Invalid)
                    lastIndex = list.Count - 1;

                for (int index = firstIndex; index <= lastIndex; index++)
                {
                    if (list[index] != null)
                        range.Add(list[index].ToString());
                    else
                        range.Add(null);
                }
            }

            return range;
        }

        ///////////////////////////////////////////////////////////////////////

        public static StringList GetRangeAsStringList(
            ArgumentList list,
            int firstIndex
            )
        {
            return GetRangeAsStringList(list, firstIndex,
                (list != null) ? (list.Count - 1) : Index.Invalid);
        }

        ///////////////////////////////////////////////////////////////////////

        public static StringList GetRangeAsStringList(
            IList list,
            int firstIndex,
            int lastIndex
            )
        {
            StringList range = null;

            if (list != null)
            {
                range = new StringList();

                if (firstIndex == Index.Invalid)
                    firstIndex = 0;

                if (lastIndex == Index.Invalid)
                    lastIndex = list.Count - 1;

                for (int index = firstIndex; index <= lastIndex; index++)
                {
                    if (list[index] != null)
                        range.Add(list[index].ToString());
                    else
                        range.Add((string)null);
                }
            }

            return range;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Public Methods
        public void AddRange(
            IEnumerable<string> collection
            )
        {
            foreach (string item in collection)
                this.Add(item);
        }

        ///////////////////////////////////////////////////////////////////////

        public bool IsVariadic(
            bool setFlags
            )
        {
            //
            // NOTE: Grab the count as we need to use it several
            //       times in this method.
            //
            int count = this.Count;

            //
            // NOTE: Does this argument list accept a variable
            //       numbers of arguments (COMPAT: Tcl)?
            //
            bool result = (count > 0) && (String.Compare(
                this[count - 1].Name, TclVars.Arguments,
                StringOps.SystemStringComparisonType) == 0);

            //
            // HACK: Flag the final argument as an "argument list"
            //       if necessary so that the syntax error message,
            //       if necessary, will be correct.
            //
            if (result)
                this[count - 1].Flags |= ArgumentFlags.ArgumentList;

            return result;
        }

        ///////////////////////////////////////////////////////////////////////

        public int GetOptionalCount()
        {
            //
            // NOTE: Return the total number of arguments in the list
            //       that are optional.  In order for an argument to
            //       be considered optional, it must meet the following
            //       criteria:
            //
            //       1. It must have a default value that is not null.
            //
            //       2. No non-optional arguments may occur after it
            //          in the argument list (COMPAT: Tcl).
            //
            int result = 0;

            //
            // NOTE: Grab the count as we need to use it several times in
            //       this method.
            //
            int count = this.Count;

            //
            // NOTE: Count all the arguments starting from the end of
            //       the list going backward that have a default value.
            //
            int index = IsVariadic(false) ? count - 2 : count - 1;

            for (; index >= 0; index--)
            {
                Argument argument = this[index];

                if ((argument != null) &&
                    FlagOps.HasFlags(argument.Flags,
                        ArgumentFlags.HasDefault, true))
                {
                    result++;
                }
                else
                {
                    break;
                }
            }

            return result;
        }

        ///////////////////////////////////////////////////////////////////////

        public string ToRawString()
        {
            StringBuilder result = StringOps.NewStringBuilder();

            foreach (Argument element in this)
                result.Append(element);

            return result.ToString();
        }

        ///////////////////////////////////////////////////////////////////////

        public string ToRawString(
            ToStringFlags toStringFlags,
            string separator
            )
        {
            StringBuilder result = StringOps.NewStringBuilder();

            foreach (Argument element in this)
            {
                if (element != null)
                {
                    if ((separator != null) && (result.Length > 0))
                        result.Append(separator);

                    result.Append(element.ToString(toStringFlags));
                }
            }

            return result.ToString();
        }

        ///////////////////////////////////////////////////////////////////////

        public string ToString(
            ToStringFlags toStringFlags
            )
        {
            return ToString(toStringFlags, null, false);
        }

        ///////////////////////////////////////////////////////////////////////

        public string ToString(
            ToStringFlags toStringFlags,
            string pattern,
            bool noCase
            )
        {
            return GenericOps<Argument>.ListToString(
                this, Index.Invalid, Index.Invalid, toStringFlags,
                Characters.Space.ToString(), pattern, noCase);
        }

        ///////////////////////////////////////////////////////////////////////

        public string ToString(
            string separator,
            string pattern,
            bool noCase
            )
        {
            return GenericOps<Argument>.ListToString(
                this, Index.Invalid, Index.Invalid, ToStringFlags.None,
                separator, pattern, noCase);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region System.Object Overrides
        public override string ToString()
        {
            return ToString(Characters.Space.ToString(), null, false);
        }
        #endregion
    }
}
