/*
 * TclForm.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using System.Collections.Generic;
using System.Threading;
using System.Windows.Forms;
using Eagle._Attributes;
using Eagle._Components.Public;

namespace TclSample
{
    //
    // FIXME: Always change this GUID.
    //
    [ObjectId("7d8ce99d-5445-4852-b752-dd2519c3bc47")]
    public partial class TclForm : Form
    {
        #region Private Constants
        private const CreateFlags DefaultCreateFlags =
            CreateFlags.Initialize | CreateFlags.ThrowOnDisposed;

        private const CommandFlags DefaultCommandFlags = CommandFlags.None;

        private const string ClosingError =
            "Please cancel the running script(s) prior to closing this form.";
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Private Tcl Loader Constants
        //
        // NOTE: What mechanism(s) do we want to use to find the
        //       versions of Tcl installed on this machine?
        //
        private const FindFlags DefaultFindFlags = FindFlags.All;

        //
        // NOTE: What is the Tcl version should we assume when the
        //       version cannot be deduced from the file name?
        //
        private static readonly Version DefaultTclVersion = new Version(8, 4);

        //
        // NOTE: What is the minimum version of Tcl we are willing
        //       to accept (null means any)?
        //
        private static readonly Version MinimumTclVersion = null;

        //
        // NOTE: What is the maximum version of Tcl we are willing
        //       to accept (null means any)?
        //
        private static readonly Version MaximumTclVersion = null;
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Private Static Data
        private static object syncRoot = new object();

        private static Dictionary<TclForm, object> forms =
            new Dictionary<TclForm, object>();
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Private Data
        private string[] args;
        private Interpreter interpreter;
        private string interpName;
        private bool disposed;
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Private Constructors
        private TclForm()
            : base()
        {
            InitializeComponent();

            //
            // NOTE: Setup the form events that we care about.
            //
            this.FormClosing +=
                new FormClosingEventHandler(TclForm_FormClosing);

            this.Disposed += new EventHandler(TclForm_Disposed);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Static "Factory" Members
        public static TclForm Create(
            string[] args
            )
        {
            TclForm form = null;
            Interpreter interpreter = null;
            ReturnCode code = ReturnCode.Ok;
            Result result = null;

            try
            {
                form = new TclForm();

                //
                // NOTE: Save the passed in command line arguments for
                //       later.
                //
                form.args = args;

                //
                // NOTE: Create a new Eagle interpreter; currently, this
                //       is necessary to use the Tcl integration features
                //       of Eagle.
                //
                interpreter = Interpreter.Create(args, DefaultCreateFlags,
                    ref result);

                if (interpreter != null)
                {
                    //
                    // NOTE: Automatically locate and select the "best"
                    //       available build of Tcl, if any.
                    //
                    code = interpreter.LoadTcl(DefaultFindFlags, null,
                        MinimumTclVersion, MaximumTclVersion,
                        DefaultTclVersion, null, ref result);

                    if (code == ReturnCode.Ok)
                    {
                        //
                        // NOTE: Grab the name for the newly created
                        //       Tcl interpreter.
                        //
                        string interpName = result;

                        //
                        // NOTE: Create a new command object.  This will
                        //       be used to demonstrate implementing a
                        //       Tcl command in managed code.
                        //
                        Class6 class6 = new Class6(new CommandData(
                            "class6", null, null, new ClientData(form),
                            typeof(Class6).FullName, DefaultCommandFlags,
                            null, 0));

                        //
                        // NOTE: Create a Tcl command bridging object.
                        //       This is used to translate inbound
                        //       native calls for a particular command
                        //       to a managed method call.
                        //
                        code = interpreter.AddTclBridge(class6,
                            interpName, class6.Name, class6.ClientData,
                            false, false, ref result);

                        //
                        // NOTE: Did we successfully create the bridged
                        //       command?
                        //
                        if (code == ReturnCode.Ok)
                        {
                            //
                            // NOTE: Save the created Eagle interpreter.
                            //
                            form.interpreter = interpreter;

                            //
                            // NOTE: Save the name of the created Tcl
                            //       interpreter.
                            //
                            form.interpName = interpName;

                            //
                            // NOTE: Keep track of this form in the
                            //       master collection.  This collection
                            //       is primarily used when script
                            //       cancellation needs to be performed
                            //       for every Tcl interpreter we have
                            //       created.
                            //
                            lock (syncRoot)
                            {
                                if (forms != null)
                                    forms.Add(form, null);
                            }
                        }
                    }
                }
                else
                {
                    code = ReturnCode.Error;
                }
            }
            catch (Exception e)
            {
                result = e;
                code = ReturnCode.Error;
            }
            finally
            {
                if (code != ReturnCode.Ok)
                {
                    //
                    // NOTE: We failed to fully create the form.  All
                    //       the resources we managed to successfully
                    //       create must be cleaned up now.
                    //
                    if (interpreter != null)
                    {
                        //
                        // NOTE: Dispose of the Eagle interpreter now
                        //       just in case the form does not have
                        //       the private variable set correctly.
                        //
                        interpreter.Dispose(); /* throw */
                        interpreter = null;
                    }

                    if (form != null)
                    {
                        //
                        // NOTE: If the form had an Eagle interpreter,
                        //       it should have already been disposed
                        //       (above) because we use the local
                        //       variable prior to setting the form
                        //       variable.  To be sure that the now
                        //       disposed Eagle interpreter is not
                        //       used during form disposal, we need to
                        //       null out the form variable now.
                        //
                        form.interpreter = null;

                        //
                        // NOTE: Now, dispose any other resources held
                        //       by the form.
                        //
                        form.Dispose();
                        form = null;
                    }
                }
            }

            if (code != ReturnCode.Ok)
                ShowResult(code, result);

            return form;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Form Event Handlers
        private void TclForm_FormClosing(
            object sender,
            FormClosingEventArgs e
            )
        {
            if (!disposed)
            {
                //
                // NOTE: Prevent this instance from being closed if
                //       a script is active.
                //
                if ((interpreter != null) &&
                    interpreter.IsTclInterpreterActive(interpName))
                {
                    e.Cancel = true;

                    ShowResult(ReturnCode.Error, ClosingError);
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////

        private void TclForm_Disposed(
            object sender,
            EventArgs e
            )
        {
            if (!disposed)
            {
                if (interpreter != null)
                {
                    //
                    // NOTE: Unload the Tcl library and cleanup
                    //       the Tcl objects that we created.
                    //       This is required because it holds
                    //       native resources.  Technically, this
                    //       step is optional if you dispose the
                    //       interpreter (below); however, it is
                    //       good practice.
                    //
                    ReturnCode unloadCode;
                    Result unloadError = null;

                    unloadCode = interpreter.UnloadTcl(
                        UnloadFlags.Default, ref unloadError);

                    if (unloadCode != ReturnCode.Ok)
                        ShowResult(unloadCode, unloadError);

                    //
                    // NOTE: Dispose of the Eagle interpreter now.
                    //       This is required because it may hold
                    //       native resources.
                    //
                    if (interpreter != null)
                    {
                        interpreter.Dispose(); /* throw */
                        interpreter = null;
                    }
                }

                //
                // NOTE: Remove this form from the master
                //       collection.
                //
                lock (syncRoot)
                {
                    if (forms != null)
                        forms.Remove(this);
                }

                //
                // NOTE: This form is now disposed.
                //
                disposed = true;
            }
        }

        ///////////////////////////////////////////////////////////////////////

        private void btnNew_Click(
            object sender,
            EventArgs e
            )
        {
            if (!disposed)
            {
                //
                // NOTE: Create a new thread.  The new thread will create
                //       another instance of this form.
                //
                Thread thread = Engine.CreateThread(
                    interpreter, CreateFormThreadStart, 0, true, false);

                if (thread != null)
                {
                    thread.Name = String.Format(
                        "tclFormThread: {0}", interpreter);

                    thread.Start(args);
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////

        private void btnEvaluate_Click(
            object sender,
            EventArgs e
            )
        {
            if (!disposed && (interpreter != null))
            {
                //
                // NOTE: Evaluate the contents of the script text box and
                //       place the formatted results into the result text
                //       box.
                //
                ReturnCode code;
                Result result = null;

                code = interpreter.EvaluateTclScript(interpName,
                    txtScript.Text, ref result);

                txtResult.Text = Utility.FormatResult(code, result,
                    interpreter.GetTclInterpreterErrorLine(interpName));
            }
        }

        ///////////////////////////////////////////////////////////////////////

        private void btnCancel_Click(
            object sender,
            EventArgs e
            )
        {
            if (!disposed)
            {
                //
                // NOTE: Cancel the active script, if any, for every form
                //       instance we know about.  Please note that we could
                //       surgically cancel the active script for a particular
                //       form instance here also; however, some kind of user
                //       interface to select a particular form instance would
                //       be required in that case.
                //
                ReturnCode code;
                Result error = null;

                code = CancelAll(false, ref error);

                if (code != ReturnCode.Ok)
                    ShowResult(code, error);
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Static Forms Helpers
        private static void ShowResult(
            ReturnCode code,
            Result result
            )
        {
            MessageBox.Show(
                Utility.FormatResult(code, result),
                Application.ProductName, MessageBoxButtons.OK,
                Utility.IsSuccess(code, false) ?
                MessageBoxIcon.Information : MessageBoxIcon.Error);
        }

        ///////////////////////////////////////////////////////////////////////

        private static ReturnCode CancelAll(
            bool strict,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (forms != null)
                {
                    try
                    {
                        foreach (TclForm form in forms.Keys)
                        {
                            if (form == null) // NOTE: Redundant?
                                continue;

                            //
                            // NOTE: Grab the Eagle interpreter for this
                            //       form.
                            //
                            Interpreter interpreter = form.interpreter;

                            if (interpreter == null)
                                continue;

                            //
                            // NOTE: Grab the Tcl interpreter name for
                            //       this form.
                            //
                            string interpName = form.interpName;

                            //
                            // NOTE: Cancel the script being evaluated
                            //       for this form instance, if any.
                            //
                            ReturnCode cancelCode;
                            Result cancelError = null;

                            cancelCode = interpreter.CancelTclEvaluate(
                                interpName, true, null, ref cancelError);

                            if (cancelCode != ReturnCode.Ok)
                            {
                                if (strict)
                                {
                                    //
                                    // NOTE: Strict mode means that we
                                    //       stop upon encountering an
                                    //       error and return that error
                                    //       to the caller.
                                    //
                                    error = cancelError;
                                    return cancelCode;
                                }
                                else
                                {
                                    //
                                    // NOTE: Otherwise, just keep going
                                    //       (after showing the error to
                                    //       the user).
                                    //
                                    ShowResult(cancelCode, cancelError);
                                }
                            }
                        }

                        return ReturnCode.Ok;
                    }
                    catch (Exception e)
                    {
                        error = e;
                    }
                }
                else
                {
                    error = "invalid forms collection";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        private static void CreateFormThreadStart(
            object obj
            )
        {
            try
            {
                //
                // NOTE: The data passed to this thread start routine
                //       must be a string array or null.
                //
                string[] args = obj as string[];
                TclForm form = null;

                try
                {
                    //
                    // NOTE: Create a new instance of the Tcl form, passing
                    //       the arguments we received from the other thread.
                    //
                    form = Create(args);

                    //
                    // NOTE: Upon success, show the form modally.
                    //
                    if (form != null)
                        form.ShowDialog();
                }
                finally
                {
                    //
                    // NOTE: Upon completion, successful or otherwise, dispose
                    //       of the form and its contained resources (including
                    //       the Tcl interpreter).
                    //
                    if (form != null)
                    {
                        form.Dispose();
                        form = null;
                    }
                }
            }
            catch (Exception e)
            {
                ShowResult(ReturnCode.Error, e);
            }
        }
        #endregion
    }
}
